/* 
*  This file is a collection of register level E-Series functions
*  Included in this file are:
*							 DAQ_STC_Windowed_Mode_Write
*							 DAQ_STC_Windowed_Mode_Read
*							 Board_Write
*							 Board_Read
*/ 

#include "eserrlp.h"

/* This function implements the STC windowed writing mode.  First *
 * the register address is written to the Window_Address_Register.*
 * The data is then written to the Window_Data_Write_Register.    *
 * Both writes are unsigned 16 bit.                               */
 
void DAQ_STC_Windowed_Mode_Write(unsigned uRegister_Address, unsigned uData)
{
	_outpw(Board_Base_Address + Window_Address_Register * 2,uRegister_Address);
	_outpw(Board_Base_Address + Window_Data_Write_Register * 2,uData); 
}

/* This function implements the STC windowed reading mode.  First *
 * the register address is written to the Window_Address_Register.*
 * The data is then read from the Window_Data_Write_Register.     *
 * Both the read and write are unsigned 16 bit.                   */

unsigned DAQ_STC_Windowed_Mode_Read(unsigned uRegister_Address)
{
	_outpw(Board_Base_Address + Window_Address_Register * 2,uRegister_Address);
	return(_inpw(Board_Base_Address + Window_Data_Read_Register * 2)); 
}

/* This function writes to an E-Series board register             */
 
void Board_Write(unsigned uRegister_Address, unsigned uData)
{
	_outpw(Board_Base_Address + uRegister_Address,uData);
}

/* This function reads from an E-Series board register            */

unsigned Board_Read(unsigned uRegister_Address)
{
	return(_inpw(Board_Base_Address + uRegister_Address));
}


