#include <linux/config.h>
#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/dma.h>

#include <stdio.h>
#include <stdarg.h>
#include <linux/delay.h> /* for usleep */

#include "ioctl_defs.h"
#include "mio64e3.h"
#include "bitdef.h"
#include "io_port_MS-C.h"
#include "eserfnct.c"
#include "mio64e3common.c"

static int               mio64e3_err;

char                     kernel_version[]=UTS_RELEASE;
struct wait_queue      * mio64e3_wait_queue;
short                    irq_A_channel;
short                    irq_B_channel;
unsigned short           dma_size=0;

unsigned short           end_of_scan=1;
unsigned short           fifo_full=0;
  
unsigned short           I_have_channels=0;
unsigned short           I_have_scans=0;
unsigned long            number_of_sequences=-EINVAL;
short                    number_of_channels=-EINVAL;
short                    scan_size=0;
short                    somebody_is_reading=0;

unsigned short           last_load_register=A;
unsigned short           last_sequence=1;
unsigned short           debug_value=0;

double                 * mcopy_d;

struct list_channel_s  * cp=NULL; /* channel pointer */
struct list_sequence_s * sp=NULL; /* sequence pointer */
struct channels_s        cps={NULL,NULL}; /* channel pointer structure */
struct sequences_s       sps={NULL,NULL}; /* sequence pointer structure */

struct list_buf_s{
   unsigned short dma_size; /* dma_size counts the scans that go in one buffer */
   char               *buf, /* starting adress of the buffer                   */
                  *bufread,	/* position where the user reads from the buffer   */
                 *bufwrite; /* position where the driver writes to the buffer  */
   struct list_buf_s *next;};

struct bufs_s{
   unsigned int c;		/* number of bytes in all buffers that can be retrived */
   struct list_buf_s *f;	/* pointer to the first buffer    */
   struct list_buf_s *l;};	/* pointer to the last buffer     */

struct bufs_s b={0,NULL,NULL};

void free_buf(){
   struct list_buf_s *t;
   kfree_s((void*)(*b.f).buf,(*b.f).dma_size*scan_size);
   t=b.f;
   b.f=(*(b.f)).next;
   if (b.f==NULL){
      b.l=NULL;}
   kfree_s((void*)t,sizeof(struct list_buf_s));}

void delete_buf(){
   disable_dma(dma_channel);
   while(b.f!=NULL){
      free_buf();}
   b.c=0;} 

int add_buf(){
   struct list_buf_s *t;
   if ((t=(struct list_buf_s*)kmalloc(sizeof(struct list_buf_s), GFP_ATOMIC))==NULL){
      printk( "mio64e3: Unable to allocate new list element.\n" );
      return -ENOMEM;}
   if (((*t).buf=(char*)kmalloc(dma_size*scan_size, GFP_DMA|GFP_ATOMIC))==NULL){
      printk( "mio64e3: Unable to allocate new DMA buffer\n" );
      return -ENOMEM;}
   (*t).dma_size=dma_size;
   (*t).bufwrite=(*t).buf;
   (*t).bufread=(*t).buf;
   (*t).next=NULL;
   if (b.l==NULL){
      b.l=b.f=t;}
   else{
      (*b.l).next=t;
      b.l=t;}
   return 0;}

unsigned long assign_scan_time(double scan_interval){
   return (unsigned long)(scan_interval*clock_Hz-1.);}

unsigned short assign_dma_size(double scan_interval){
/* The dma_size is chosen relative to the scan_time definition. */
/* The wake_up_interruptible in mio64e3_interrupt will be done  */
/* each 'update_interval' seconds.                              */
   unsigned short i;
   if ((*sps.fsp).count==1){ /* single scan mode */
      return 1;}
   i=(unsigned short)(update_interval/scan_interval); /* Division by zero inhibit in */
						      /* Number_Of_Sequences.        */
   if (i<=0) return -EINVAL;
   if (i<1) return 1;
   return i;}

short Number_Of_Channels(struct list_channel_s *l){
   unsigned short t=0;
   while(l!=NULL){
      t++;
      if(t<=Max_Channel){ /* Not more than 64 channels for mio64e3! */
         l=(*l).next;}
      else{
         return -EINVAL;}}
   return t;}

int Number_Of_Sequences(struct list_sequence_s *l){
   unsigned short t=0;
   while(l!=NULL){
      t++;
      if( ((*l).count<0)          /* SC counter size limits. 0 gives continuous scanning. */
        ||((*l).count==1)         /* Mio64e3 cannot generate sequences with one scan!     */
        ||((*l).count>=(1<<24))){ /* SC counter size limits.                              */
         return -EINVAL;}
      if (  (0>assign_scan_time((*l).scan_interval))          /* SI counter size limit.        */
							      /* And implizite,                */
							      /* (*l).scan_interval>0,         */
							      /* which inhibits division       */
							      /* by zero in assign_dma_size(). */
          ||((1<<24)<=assign_scan_time((*l).scan_interval))){ /* SI counter size limit.        */
         return -EINVAL;}
      l=(*l).next;}
   return t;}


/*
* E-series helpers for Calibration Function
*
* For use with AT-MIO-16E-1, ..16E-2, and ..64E-3 only!
* You have to adapt WriteCaldac() for other E-series boards.
*/

void clock_in(short bit0){
   mirror_Serial_Command&=~SerClk;
   mirror_Serial_Command&=~SerData;
   mirror_Serial_Command|=((bit0<<1)&SerData); /* Bit position of SerData-bit is 1. */
   Board_Write(Serial_Command,mirror_Serial_Command);
   mirror_Serial_Command|=SerClk;
   Board_Write(Serial_Command,mirror_Serial_Command);
   mirror_Serial_Command&=~SerClk;
   Board_Write(Serial_Command,mirror_Serial_Command);
  }

short clock_out(){
short t;
   mirror_Serial_Command|=SerClk;
   Board_Write(Serial_Command,mirror_Serial_Command);
   t=Board_Read(Status)&PROMOUT;
   mirror_Serial_Command&=~SerClk;
   Board_Write(Serial_Command,mirror_Serial_Command);
   return t;
  }

unsigned short ReadEeprom(unsigned short adress){
short i;
unsigned short      ni_daq_board_code;
/* load from EEProm */
   mirror_Serial_Command=0;
   Board_Write(Serial_Command,mirror_Serial_Command);
   mirror_Serial_Command=EEPromCS;
   Board_Write(Serial_Command,mirror_Serial_Command);
   clock_in(0);
   clock_in(0);
   clock_in(0);
   clock_in(0);
   clock_in((adress>>8)&1);
   clock_in(0);
   clock_in(1);
   clock_in(1);
   for (i=7; i >= 0; i--){
      clock_in((adress>>i)&1);}
   ni_daq_board_code=0;
   for (i=7; i >= 0; i--){
      ni_daq_board_code=(ni_daq_board_code<<1)|clock_out();}
   Board_Write(Serial_Command,0x00);
   return ni_daq_board_code;}

/*
* This function is made for AT-MIO-16E-1, ..16E-2, and ..64E-3 only!
*/
unsigned short WriteCaldac(unsigned short adress, unsigned short cal_value){
short i;
/* write to CalDAC */
   mirror_Serial_Command=0;
   Board_Write(Serial_Command,mirror_Serial_Command);
   for (i=0; i <= 3; i++){
      clock_in((adress>>i)&1);}
   for (i=7; i >= 0; i--){
      clock_in((cal_value>>i)&1);}
   Board_Write(Serial_Command,0x00);
   mirror_Serial_Command|=SerDacLd0;
   Board_Write(Serial_Command,mirror_Serial_Command);
   mirror_Serial_Command&=~SerDacLd0;
   Board_Write(Serial_Command,mirror_Serial_Command);
   return;}


/*
* E-series Calibration Function
*
* For my purpose, just AI-calibration is needed and done.
* For AO-calibration add more ReadEeprom, WriteCaldac lines
* and uncomment the appropriate CALDAC definitions in bitdefs.h.
*/

unsigned short AI_Calibrate(){
   unsigned short caldac_val;
   caldac_val=ReadEeprom(CALDAC4);
   WriteCaldac(4,caldac_val);
   caldac_val=ReadEeprom(CALDAC3);
   WriteCaldac(3,caldac_val);
   caldac_val=ReadEeprom(CALDAC2);
   WriteCaldac(2,caldac_val);
   caldac_val=ReadEeprom(CALDAC1);
   WriteCaldac(1,caldac_val);}


/*
* E-series helpers for Main Functions 
*/

void ai_sc_load_A(unsigned long nos){
   DAQ_STC_Windowed_Mode_Write(AI_SC_Load_A_Registers,(unsigned short)((nos-1)>>16));
   DAQ_STC_Windowed_Mode_Write(AI_SC_Load_A_Registers+1,(unsigned short)((nos-1)&0x0000ffff));}

void ai_sc_load_B(unsigned long nos){
   DAQ_STC_Windowed_Mode_Write(AI_SC_Load_B_Registers,(unsigned short)((nos-1)>>16));
   DAQ_STC_Windowed_Mode_Write(AI_SC_Load_B_Registers+1,(unsigned short)((nos-1)&0x0000ffff));}

void ai_sc_load(){
   DAQ_STC_Windowed_Mode_Write(AI_Command_1_Register,AI_SC_Load);}

void ai_sc_switch_on_tc(){
   DAQ_STC_Windowed_Mode_Write(AI_Command_2_Register,AI_SC_Switch_Load_On_TC);}

void ai_si_load_A(unsigned long st){
   DAQ_STC_Windowed_Mode_Write(AI_SI_Load_A_Registers,(unsigned short)(st>>16));
   DAQ_STC_Windowed_Mode_Write(AI_SI_Load_A_Registers+1,(unsigned short)(st&0x0000ffff));}

void ai_si_load_B(unsigned long st){
   DAQ_STC_Windowed_Mode_Write(AI_SI_Load_B_Registers,(unsigned short)(st>>16));
   DAQ_STC_Windowed_Mode_Write(AI_SI_Load_B_Registers+1,(unsigned short)(st&0x0000ffff));}

void ai_si_load(){
   DAQ_STC_Windowed_Mode_Write(AI_Command_1_Register,AI_SI_Load);}

void ai_si_switch_on_sc_tc(){
   DAQ_STC_Windowed_Mode_Write(AI_Command_2_Register,AI_SI_Switch_Load_On_SC_TC);}


/*
* E-series Main Functions 
*/

void AI_Configure_Board(){
   unsigned short i;
   cp=cps.fcp;
   DAQ_STC_Windowed_Mode_Write(Write_Strobe_0_Register,Write_Strobe_0);
   DAQ_STC_Windowed_Mode_Write(Write_Strobe_1_Register,Write_Strobe_1);
   if(cp!=NULL){
      while(((*cp).next)!=NULL){
         Board_Write(Configuration_Memory_High,(*cp).cmh);
         Board_Write(Configuration_Memory_Low,(*cp).cml);
         cp=(*cp).next;}
      Board_Write(Configuration_Memory_High,(*cp).cmh);
      Board_Write(Configuration_Memory_Low,((*cp).cml)|LastChan);}
   return;}

void AI_MSC_Clock_Configure(){
   /* 100kHz time base */
   DAQ_STC_Windowed_Mode_Write(Clock_and_FOUT_Register,Slow_Internal_Time_Divide_By_2
                                                      |Slow_Internal_Timebase
                                                      |Clock_To_Board_Divide_By_2
                                                      |Clock_To_Board);
   return;}

void AI_Reset_All(){
   /* AI_Reset even disarms the counters. */
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Reset);
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_Start);
   DAQ_STC_Windowed_Mode_Write(Interrupt_A_Enable_Register,0x0000);
   DAQ_STC_Windowed_Mode_Write(Second_Irq_A_Enable_Register,0x0000);
   DAQ_STC_Windowed_Mode_Write(Interrupt_B_Enable_Register,0x0000);
   DAQ_STC_Windowed_Mode_Write(Second_Irq_B_Enable_Register,0x0000);
   DAQ_STC_Windowed_Mode_Write(Interrupt_A_Ack_Register,AI_All_Interrupt_A_Ack
                                                       |AI_All_Error_A_Confirm);
   DAQ_STC_Windowed_Mode_Write(Interrupt_B_Ack_Register,AI_All_Interrupt_B_Ack
                                                       |AI_All_Error_B_Confirm);
   Board_Write(AI_AO_Select_Register,Input_DMA_A);
   Board_Write(Channel_A_Mode_Register,dma_channel); /* turn off DMA A */
   Board_Write(Strobes_Register,DmaTcAClr); /* clears the DMA A IRQ */
   /* disable IRQ's but do set the good lines. */
   mirror_Interrupt_Control_Register=irq_A_channel /* turn of IRQ A and B */
                                    |irq_B_channel
                                    |Pass_Thru_0_Interrupt_Polarity; /* just turn of the stauts flag */
   DAQ_STC_Windowed_Mode_Write(Interrupt_Control_Register,mirror_Interrupt_Control_Register);
   mirror_AI_Mode_1_Register=AI_Start_Stop
                            |Reserved_One;
   DAQ_STC_Windowed_Mode_Write(AI_Mode_1_Register,mirror_AI_Mode_1_Register);
   mirror_AI_Mode_2_Register=0;
   mirror_AI_Mode_3_Register=0;
   mirror_AI_START_STOP_Select_Register=0;
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_End);
   return;}

void AI_Board_Personalize(){
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_Start);
   DAQ_STC_Windowed_Mode_Write(AI_Personal_Register,AI_SHIFTIN_Pulse_Width
                                                   |AI_SOC_Polarity
                                                   |AI_CONVERT_Pulse_Width
                                                   |AI_Overrun_Mode
                                                   |AI_LOCALMUX_CLK_Pulse_Width); /* 0xA4A0 */
   /* Select Output polarities. */
   DAQ_STC_Windowed_Mode_Write(AI_Output_Control_Register,0x072E);
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_End);
   return;}

void AI_Initialize_Configuration_Memory_Output(){
/* This does not work according to Technical Reference Manual 2-21: !            */
/* You do not get the CONVERT_Pulse!                                             */
/*     DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_Start); */

     DAQ_STC_Windowed_Mode_Write(AI_Command_1_Register,AI_CONVERT_Pulse);

/* dito:                                                                         */
/*     DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_End);   */
/*                                                                               */
     return;}

void AI_Board_Environmentalize(){
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_Start);
   /* Generate a pulse on       */
   /* PFI7 on each START signal.*/
   /* It is used by the pcl833  */
   /* board as IRQ-trigger.     */
   DAQ_STC_Windowed_Mode_Write(IO_Bidirection_Pin_Register,BD_7_Pin_Dir);
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_End);
   return;}

void AI_FIFO_Request(){
   /* AI_Mode_3_Register:                            */
   /* AI_FIFO_Mode = 0, IRQ or DMA on FIFO not empty */
   return;}

void AI_Trigger_Signals(){
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_Start);
   DAQ_STC_Windowed_Mode_Write(AI_Trigger_Select_Register,AI_START1_Edge
                                                         |AI_START1_Sync);
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_End);
   return;}

void AI_Number_of_Scans__AI_Scan_Start(){
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_Start);

   mirror_AI_Mode_1_Register|=AI_Continuous;
   DAQ_STC_Windowed_Mode_Write(AI_Mode_1_Register,mirror_AI_Mode_1_Register);

   /* AI_Number_of_Scans */
   sp=sps.fsp;
   ai_sc_load_A((*sp).count);
   ai_sc_load();

   /* AI_Scan_Start */
   mirror_AI_START_STOP_Select_Register|=(AI_START_Select_SI_TC
                                         |AI_START_Edge
                                         |AI_START_Sync);
   mirror_AI_Mode_1_Register|=AI_SI_Source_IN_TIMEBASE2;
   DAQ_STC_Windowed_Mode_Write(AI_START_STOP_Select_Register,mirror_AI_START_STOP_Select_Register);
   DAQ_STC_Windowed_Mode_Write(AI_Mode_1_Register,mirror_AI_Mode_1_Register);
   ai_si_load_A(assign_scan_time((*sp).scan_interval));
   ai_si_load();
   last_load_register=A;

   last_sequence=1;
   if (((*sp).next)!=NULL){     /* Let the mio-board auto-reload,            */
                                /* we have to prepare for the next scan too. */
      last_sequence=0;
      sp=(*sp).next;

   /* AI_Number_of_Scans */
      ai_sc_load_B((*sp).count);
      ai_sc_switch_on_tc();

   /* AI_Scan_Start */
      ai_si_load_B(assign_scan_time((*sp).scan_interval));
      ai_si_switch_on_sc_tc();
      last_load_register=B;}
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_End);
   return;}

void AI_Scan_End(){
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_Start);
   mirror_AI_START_STOP_Select_Register|=AI_STOP_Sync;
   mirror_AI_START_STOP_Select_Register|=AI_STOP_Select_STOP_IN_pin;
   DAQ_STC_Windowed_Mode_Write(AI_START_STOP_Select_Register,mirror_AI_START_STOP_Select_Register);
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_End);
   return;}

void AI_Convert_Signal(){
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_Start);
   /* These Mode_2_Reg settings imply:                      */
   /* AI_SI_Initial_Load_Source = 0  =>  SI Load Register A */
   /* AI_SC_Initial_Load_Source = 0  =>  SC Load Register A */
   mirror_AI_Mode_2_Register|=AI_SI2_Reload_Mode;
   mirror_AI_Mode_3_Register|=AI_SI2_Source_Select; /* AI_IN_TIMEBASE1 */
   DAQ_STC_Windowed_Mode_Write(AI_Mode_2_Register,mirror_AI_Mode_2_Register);
   DAQ_STC_Windowed_Mode_Write(AI_Mode_3_Register,mirror_AI_Mode_3_Register);
   DAQ_STC_Windowed_Mode_Write(AI_SI2_Load_A_Register,1999);
   DAQ_STC_Windowed_Mode_Write(AI_SI2_Load_B_Register,1999);
   DAQ_STC_Windowed_Mode_Write(AI_Command_1_Register,AI_SI2_Load);
   DAQ_STC_Windowed_Mode_Write(Joint_Reset_Register,AI_Configuration_End);
   return;}

void AI_Clear_FIFO(){
   DAQ_STC_Windowed_Mode_Write(Write_Strobe_1_Register,0x0001);
   return;}

void Program_DMA_Controller(){
   add_buf();
   disable_dma(dma_channel);
   clear_dma_ff(dma_channel);
   set_dma_addr(dma_channel, (unsigned long)(*b.l).bufwrite);
   set_dma_count(dma_channel, (*b.l).dma_size*scan_size);
   set_dma_mode(dma_channel, DMA_MODE_READ);
   enable_dma(dma_channel);
   return;}

void AI_Init_DMA_Controller(){
   Board_Write(Strobes_Register,DmaTcAClr);
   Board_Write(AI_AO_Select_Register,Input_DMA_A);
   Board_Write(Channel_A_Mode_Register,TCIntEnable
                                      |dma_channel);
   Board_Write(Strobes_Register,DmaTcAClr);
   Board_Write(Channel_A_Mode_Register,ChanEnable
                                      |TCIntEnable
                                      |dma_channel);
   Board_Write(Strobes_Register,DmaTcAClr);
   return;}

void AI_Interrupt_Enable(){
   DAQ_STC_Windowed_Mode_Write(Interrupt_A_Ack_Register,AI_All_Interrupt_A_Ack
                                                       |AI_All_Error_A_Confirm);
   DAQ_STC_Windowed_Mode_Write(Interrupt_A_Ack_Register,AI_All_Interrupt_B_Ack
                                                       |AI_All_Error_B_Confirm);
   DAQ_STC_Windowed_Mode_Write(Interrupt_A_Enable_Register,AI_SC_TC_Interrupt_Enable);
   DAQ_STC_Windowed_Mode_Write(Second_Irq_A_Enable_Register,0x0000);
   DAQ_STC_Windowed_Mode_Write(Interrupt_B_Enable_Register,Pass_Thru_1_Interrupt_Enable);
   DAQ_STC_Windowed_Mode_Write(Second_Irq_B_Enable_Register,0x0000);
   DAQ_STC_Windowed_Mode_Write(Interrupt_Control_Register,mirror_Interrupt_Control_Register);
   DAQ_STC_Windowed_Mode_Write(Interrupt_Control_Register,mirror_Interrupt_Control_Register
                                                         |Interrupt_A_Enable
                                                         |Interrupt_B_Enable);
  return;}

void AI_Arming(){
   /* The arming bitfield must be set in a single write! */
   DAQ_STC_Windowed_Mode_Write(AI_Command_1_Register,AI_SI2_Arm|AI_SI_Arm|AI_SC_Arm);
   return;}

void AI_Start_The_Acquisition(){
   if((*sps.fsp).count==1){ /* Single scan exception */
      /* The AI_End_On_... bits must be set later than the arming of the counters is done! */
      DAQ_STC_Windowed_Mode_Write(AI_Command_2_Register,AI_End_On_End_Of_Scan);}
   DAQ_STC_Windowed_Mode_Write(AI_Command_2_Register,AI_START1_Pulse);
   return;}

void AI_Exit(){
   AI_Reset_All();
   disable_dma(dma_channel);
   end_of_scan=1;
   return;}

void AI_ISR(){
   if (last_sequence){
      if  (((*sp).count)==0){                  /* Do continuous reading */
         Program_DMA_Controller();}
      else{
         AI_Exit();}}
   else{
      dma_size=assign_dma_size((*sp).scan_interval);
      Program_DMA_Controller();
      if (((*sp).next)==NULL){
  	 if (((*sp).count)!=6){
            DAQ_STC_Windowed_Mode_Write(AI_Command_2_Register,AI_End_On_SC_TC);}
         last_sequence=1;}
      else {
         sp=(*sp).next;
         switch (last_load_register){
         case A:   ai_sc_load_B((*sp).count);
         		ai_si_load_B(assign_scan_time((*sp).scan_interval));
         		last_load_register=B;
   		break;
         case B:   ai_sc_load_A((*sp).count);
         		ai_si_load_A(assign_scan_time((*sp).scan_interval));
         		last_load_register=A;
   		break;}
         ai_sc_switch_on_tc();
         ai_si_switch_on_sc_tc();}}
   return;}

static int mio64e3_open(struct inode * inode, struct file * file){
   MOD_INC_USE_COUNT;
   return 0;}

static int mio64e3_read(struct inode * inode, struct file * file,
                        char * buffer, int counts){
   unsigned int j;
   somebody_is_reading=1;
   while ((b.c<counts)&&(!end_of_scan))
      interruptible_sleep_on (&mio64e3_wait_queue);
   somebody_is_reading=0;
   if (fifo_full){
      counts=FIFO_full;}
   else {
      if (b.c<counts){
         counts=b.c;}
      for(j=0;counts>j;j++){
         if (b.f!=NULL){
            if ((*b.f).bufread!=(*b.f).bufwrite){
               put_fs_byte(*(*b.f).bufread,buffer);
               (*b.f).bufread++;
               b.c--;
               buffer++;
               if ((*b.f).bufread==(*b.f).bufwrite){
                  free_buf();}}}}}
   return counts;}

static void mio64e3_interrupt (){
   unsigned short TC_status; 
   unsigned short AI_status; 
   unsigned short AO_status; 
   cli();
   TC_status=(unsigned short)Board_Read(Status)&DMATCA;
   AI_status=DAQ_STC_Windowed_Mode_Read(AI_Status_1_Register)&AI_SC_TC_St;
   AO_status=DAQ_STC_Windowed_Mode_Read(AO_Status_1_Register)&(Interrupt_B_St|Pass_Thru_1_Interrupt_St);

   if (AI_Overflow_St&DAQ_STC_Windowed_Mode_Read(AI_Status_1_Register)){
      fifo_full=1;}

/* DMA buffer full, SC_TC no */
   if ((TC_status==DMATCA)&&
       (AI_status!=AI_SC_TC_St)&&
       (AO_status==(Interrupt_B_St|Pass_Thru_1_Interrupt_St))){
      b.c+=(*b.l).dma_size*scan_size;
      (*b.l).bufwrite+=(*b.l).dma_size*scan_size;
      Program_DMA_Controller();
      Board_Write(Strobes_Register,DmaTcAClr);        /* Clears the DMA A IRQ.       */
      if ((*sps.fsp).count==1){                       /* We do sample a single scan. */
         AI_Exit();
/*
         AI_Number_of_Scans__AI_Scan_Start();
         AI_Convert_Signal();
         AI_Arming();
*/
	 }}
/* DMA buffer full, SC_TC yes */
   if ((TC_status==DMATCA)&&
       (AI_status==AI_SC_TC_St)&&
       (AO_status==(Interrupt_B_St|Pass_Thru_1_Interrupt_St))){
      disable_dma(dma_channel);
      b.c+=(*b.l).dma_size*scan_size;
      (*b.l).bufwrite+=(*b.l).dma_size*scan_size;
      AI_ISR();
      DAQ_STC_Windowed_Mode_Write(Interrupt_A_Ack_Register,
                                  AI_SC_TC_Interrupt_Ack); /* clears the SC_TC IRQ */
      Board_Write(Strobes_Register,DmaTcAClr);}
/* DMA buffer not full, SC_TC yes */
   if ((TC_status!=DMATCA)&&(AI_status==AI_SC_TC_St)){
      disable_dma(dma_channel);
      b.c+=(*b.l).dma_size*scan_size-get_dma_residue(dma_channel);
      (*b.l).bufwrite+=(*b.l).dma_size*scan_size-get_dma_residue(dma_channel);
      AI_ISR();
      DAQ_STC_Windowed_Mode_Write(Interrupt_A_Ack_Register,AI_SC_TC_Interrupt_Ack);}
   wake_up_interruptible (&mio64e3_wait_queue);
   sti();
   }

static void mio64e3_release(struct inode * inode, struct file * file){
   MOD_DEC_USE_COUNT;}

static signed int mio64e3_ioctl(struct inode * inode, struct file * file,
			unsigned int func, unsigned long option){

/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
/* :                                                      : */
/* :   All fuctions are called  through  mio64e3()         : */
/* :                                                      : */
/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::: */

switch(func){

case Debug:
/* debug_value=ReadEeprom(option); */
   debug_value=DAQ_STC_Windowed_Mode_Read(AI_SC_Save_RegistersL)
              |DAQ_STC_Windowed_Mode_Read(AI_SC_Save_RegistersH);
/*
   DAQ_STC_Windowed_Mode_Write(AI_Command_2_Register,AI_SC_Save_Trace);
            debug_value=b.c;
            debug_value=DAQ_STC_Windowed_Mode_Read(AI_Status_1_Register)
                       |Board_Read(Status)
                       |DAQ_STC_Windowed_Mode_Read(AO_Status_1_Register);

            DAQ_STC_Windowed_Mode_Write(Interrupt_A_Ack_Register,AI_All_Interrupt_A_Ack);
   DAQ_STC_Windowed_Mode_Write(AI_Command_2_Register,AI_End_On_SC_TC
                                                    |AI_End_On_End_Of_Scan);
*/
            return debug_value;

case Get_Number_Of_Channels:
   return number_of_channels;

case Get_Number_Of_Scans:
   if (sp==NULL){
      return -EINVAL;}
   return (*sp).count;

case Get_Number_Of_Sequences:
   return number_of_sequences;

case Init:
   if (  (!I_have_channels)
       ||(!I_have_scans)
       || (!end_of_scan)){            /* Do not 'Init' while reading data. */
      return -EINVAL;}
   end_of_scan=0; /* obsolete if we keep ckecking end_of_scan two lines above */
   fifo_full=0;
   cli();
   AI_Calibrate();
   AI_Reset_All();
   AI_MSC_Clock_Configure();
   AI_Board_Personalize();
   /* Set channels                                               */
   /* This neads a CONVERT_Pulse to return to the first channel, */
   /* otherwise you get one extra CONVERT_Pulse and extra data   */
   /* after starting the acquisition with a START1_Pulse.        */
   AI_Configure_Board();
   AI_Initialize_Configuration_Memory_Output(); /* Convert Pulse */
   AI_Clear_FIFO();
   AI_Board_Environmentalize();
   AI_FIFO_Request();
   AI_Trigger_Signals();
   AI_Number_of_Scans__AI_Scan_Start();
   AI_Scan_End();
   AI_Convert_Signal();
   AI_Clear_FIFO();
   delete_buf();
   dma_size=assign_dma_size((*sps.fsp).scan_interval);
   Program_DMA_Controller();
   AI_Init_DMA_Controller();
   AI_Interrupt_Enable();
   AI_Arming();
   sti();
   break;

case Set_Channel:
   if (!end_of_scan){            /* Do not 'Set_Channel' while reading data. */
      return -EINVAL;}
   I_have_channels=0;
   delete_channels_s(&cps);
   cp=NULL;
   number_of_channels=-EINVAL;
   scan_size=0;
   if ((struct list_channel_s *)option==NULL){
      return -EINVAL;}
   cp=(struct list_channel_s *)option;
   do {
      if ((-ENOMEM)==add_channel(&cps, get_fs_word(&(*cp).cmh), get_fs_word(&(*cp).cml))){
         printk( "mio64e3: Unable to allocate memory for new channel list element.\n" );
         delete_channels_s(&cps);
         cp=NULL;
         return -ENOMEM;}
      cp=(struct list_channel_s *)get_fs_long(&(*cp).next);}
   while (cp!=NULL);
   cp=cps.fcp;
   number_of_channels=Number_Of_Channels(cps.fcp);
   if (number_of_channels==-EINVAL){        /* Might be more than Max_Channel allows. */
      delete_channels_s(&cps);
      cp=NULL;
      return -EINVAL;}
   scan_size=number_of_channels*sizeof(unsigned short);
   I_have_channels=1;
   break;

case Set_Sequence:
   if (!end_of_scan){            /* Do not 'Set_Sequence' while reading data. */
      return -EINVAL;}
   I_have_scans=0;
   delete_sequences_s(&sps);
   sp=NULL;
   if ((struct list_sequence_s *)option==NULL){
      return -EINVAL;}
   sp=(struct list_sequence_s *)option;
   do {
      memcpy_fromfs(mcopy_d, &(*sp).scan_interval, sizeof(double));
      if ((-ENOMEM)==add_sequence(&sps, *mcopy_d, get_fs_long(&(*sp).count))){
         printk( "mio64e3: Unable to allocate memory for new sequence list element.\n" );
         delete_sequences_s(&sps);
         sp=NULL;
         return -ENOMEM;}
      sp=(struct list_sequence_s *)get_fs_long(&(*sp).next);}
   while (sp!=NULL);
   sp=sps.fsp;
   if  (((*sps.fsp).count==1)
      &&((*sps.fsp).next==NULL)){
      number_of_sequences=1;}
   else{
      number_of_sequences=Number_Of_Sequences(sps.fsp); /* validation */
      if (number_of_sequences==-EINVAL){
         delete_sequences_s(&sps);
         sp=NULL;
         return -EINVAL;}}
   dma_size=assign_dma_size((*sps.fsp).scan_interval);
   I_have_scans=1;
   break;

case Start:
   if (somebody_is_reading==0){ /* Do not start the pump without having an exhaust pipe. */
      return -EINVAL;}
   AI_Start_The_Acquisition();
   break;

case Stop:
    cli();
   AI_Reset_All();
   /* clean the barrel */
   if(!end_of_scan){
      disable_dma(dma_channel);
      /* take only whole scans */
      b.c+=(((*b.l).dma_size*scan_size-get_dma_residue(dma_channel))
           /scan_size)*scan_size;
      (*b.l).bufwrite+=(((*b.l).dma_size*scan_size-get_dma_residue(dma_channel))
                       /scan_size)*scan_size;
      end_of_scan=1;
      wake_up_interruptible (&mio64e3_wait_queue);}
    sti();
   break;

default: return -EINVAL;
}
return 0;
}

static struct file_operations mio64e3_fops = {
	NULL,           /*  seek    */
	mio64e3_read,   /*  read    */
	NULL,           /*  write   */
	NULL,           /*  readdir */
	NULL,           /*  select  */
	mio64e3_ioctl,  /*  ioctl   */
	NULL,           /*  mmap    */
	mio64e3_open,   /*  open    */
	mio64e3_release /*  release */
};

#ifndef MODULE
long mio64e3_init(long mem_start, long mem_end) {
  if (0>(mio64e3_err = register_chrdev(Mio64e3_Major, "mio64e3", &mio64e3_fops)))
    printk("mio64e3: unable to get major for mio64e3 device\n");
  return mem_start;
}
#else
int init_module(void)
{
   switch (Irq_A){
   case  3: irq_A_channel=0x0000; break;
   case  4: irq_A_channel=0x0100; break;
   case  5: irq_A_channel=0x0200; break;
   case  7: irq_A_channel=0x0300; break;
   case 10: irq_A_channel=0x0400; break;
   case 11: irq_A_channel=0x0500; break;
   case 12: irq_A_channel=0x0600; break;
   case 15: irq_A_channel=0x0700; break;
   default: printk ("mio64e3: This board does not support IRQ %d!\n", (int)Irq_A);
	    return -EIO;
   }

   switch (Irq_B){
   case  3: irq_B_channel=0x0000; break;
   case  4: irq_B_channel=0x1000; break;
   case  5: irq_B_channel=0x2000; break;
   case  7: irq_B_channel=0x3000; break;
   case 10: irq_B_channel=0x4000; break;
   case 11: irq_B_channel=0x5000; break;
   case 12: irq_B_channel=0x6000; break;
   case 15: irq_B_channel=0x7000; break;
   default: printk ("mio64e3: This board does not support IRQ %d!\n", (int)Irq_B);
	    return -EIO;
   }

   if (check_region(Base,Port_Amount)){
     printk("mio64e3: Base 0x%x already in use.\n",Base);
     return -EIO; }
 
   if (check_region(Read_Port_Address,1)){
     printk("mio64e3: PNP read port location 0x%x already in use.\n",Read_Port_Address);
     return -EIO; }

   if (request_irq (Irq_A, mio64e3_interrupt, SA_INTERRUPT, "mio64e3")){
      printk ("mio64e3: Unable to use interrupt %d.\n", (int)Irq_A);
      return -EIO;}
   free_irq(Irq_A);

   if (request_dma(dma_channel, "mio64e3")){
      printk( "mio64e3: can't request DMA channel %d\n", dma_channel);
      return -EIO;}
   free_dma(dma_channel);

   if (!(mcopy_d=(double *)kmalloc(sizeof(double)))){
      return -EINVAL;}

   if (0>(mio64e3_err = register_chrdev(Mio64e3_Major, "mio64e3", &mio64e3_fops))){
      if (mio64e3_err == -EEXIST){
         printk("mio64e3: Unable to get major for mio64e3 device.\n"); }
      else {
         printk("mio64e3: Unable to register mio64e3 device.\n"); }
      return -EIO;}
   
   request_region(Base,Port_Amount,"mio64e3"); /* doing it here I don't have
						to give it back, if the the
						initialization fails. */
   request_region(Read_Port_Address,1,"mio64e3");
   request_irq (Irq_A, mio64e3_interrupt, SA_INTERRUPT, "mio64e3");
   request_dma(dma_channel, "mio64e3");
   printk("mio64e3: Vers. 0.8 01/13/1998 (Christian Fughe)   C.M.Fughe@wtb.tue.nl\n");
   printk("mio64e3:                      (Rob de Kluijver) R.d.Kluyver@ctd.tue.nl\n");
   printk("mio64e3: Driver loaded, major %d, IRQ %d, DMA %d, Base 0x%X-0x%X.\n",
          Mio64e3_Major,(int)Irq_A,dma_channel,Base,Base+Port_Amount-1);
   return 0;}
 
void cleanup_module(void){
   if (MOD_IN_USE)
      printk("mio64e3: Busy, retry later!\n");
   else {
      unregister_chrdev(Mio64e3_Major, "mio64e3");
      delete_sequences_s(&sps);
      sp=NULL;
      delete_channels_s(&cps);
      cp=NULL;
      kfree_s((void*)mcopy_d,sizeof(double));
      delete_buf();
      printk("mio64e3: buffer released\n");
      free_dma(dma_channel);
      printk("mio64e3: dma released\n");
      free_irq(Irq_A);
      printk("mio64e3: irq released\n");
      release_region(Base,Port_Amount);
      release_region(Read_Port_Address,1);
      printk("mio64e3: Device released!\n");}}
#endif
